from rest_framework.exceptions import (
    APIException,
    MethodNotAllowed,
    PermissionDenied,
)
from rest_framework.serializers import Serializer
from rest_framework.test import APIRequestFactory

from django.core.exceptions import ImproperlyConfigured
from django.test import (
    TestCase,
    override_settings,
)

from audoma.decorators import (
    AudomaActionException,
    audoma_action,
)
from audoma.drf import fields as audoma_fields
from audoma.tests.testtools import (
    create_serializer_class,
    create_view_with_custom_audoma_action,
)


class AudomaActionTestCase(TestCase):
    def setUp(self):
        super().setUp()
        self.factory = APIRequestFactory()
        serializer_base_classes = [Serializer]
        fields_config = {
            "firstname": audoma_fields.CharField(max_length=255),
            "lastname": audoma_fields.CharField(max_length=255),
            "age": audoma_fields.IntegerField(),
        }
        self.example_collect_serializer = create_serializer_class(
            fields_config=fields_config, serializer_base_classes=serializer_base_classes
        )
        self.example_result_serializer = create_serializer_class(
            fields_config=fields_config, serializer_base_classes=serializer_base_classes
        )
        self.request_data = {"firstname": "John", "lastname": "DOG", "age": 45}

    @override_settings(DEBUG=False)
    def test_improperly_configured_audoma_action_decorator(self):
        try:
            audoma_action(
                collectors={"post": self.example_collect_serializer},
                results={"201": self.example_result_serializer},
                methods=["GET"],
                detail=True,
            )
        except ImproperlyConfigured as e:
            self.assertEqual(
                str(e),
                "There should be no collectors defined if there are not create/update requests accepted.",
            )

    def test_audoma_action_with_wrong_clas_passed_to_errors(self):
        class NotAnException:
            ...

        error = NotAnException

        try:
            audoma_action(
                collectors={"post": self.example_collect_serializer},
                results={"201": self.example_result_serializer},
                errors=[error],
                methods=["POST"],
                detail=False,
            )
        except ImproperlyConfigured as e:
            self.assertEqual(
                str(e),
                f"Something that is not an Exception instance or class has been passed \
                        to audoma_action errors list. The value which caused exception: {error}",
            )

    def test_audoma_action_multiple_exception_definitions(self):
        error = ValueError

        try:
            audoma_action(
                collectors={"post": self.example_collect_serializer},
                results={"201": self.example_result_serializer},
                errors=[error, error("Some Message")],
                methods=["POST"],
                detail=False,
            )
        except ImproperlyConfigured as e:
            self.assertEqual(
                str(e),
                f"Exception has been passed multiple times as an instance and as type, \
                        exception type: {error}",
            )

    def test_properly_defined_audoma_action(self):
        action = audoma_action(
            collectors={"post": self.example_collect_serializer},
            results={"201": self.example_result_serializer},
            methods=["POST"],
            detail=False,
        )
        self.assertEqual(type(action), audoma_action)

    def test_audoma_action_return_defined_response(self):
        request = self.factory.post("/custom_action/")
        request.data = self.request_data
        view = create_view_with_custom_audoma_action(
            request=request,
            action_kwargs={
                "collectors": {"post": self.example_collect_serializer},
                "results": {201: self.example_result_serializer},
                "methods": ["POST"],
                "detail": False,
            },
            returnables=(self.request_data, 201),
        )
        view.method = "post"
        view.format_kwarg = "json"
        view.request = request
        view.action = "custom_action"
        response = view.custom_action(request)
        self.assertEqual(response.status_code, 201)
        self.assertEqual(
            response.data, {"firstname": "John", "lastname": "DOG", "age": 45}
        )

    def test_audoma_action_return_undefined_response(self):
        request = self.factory.get("/custom_action/")
        fields_config = {"company_name": audoma_fields.CharField(max_length=255)}
        serializer_class = create_serializer_class(
            fields_config=fields_config, serializer_base_classes=[Serializer]
        )

        view = create_view_with_custom_audoma_action(
            request=request,
            action_kwargs={
                "collectors": {"post": self.example_collect_serializer},
                "results": {201: self.example_result_serializer},
                "methods": [
                    "POST",
                ],
                "detail": False,
            },
            returnables=({"company_name": "Lenovo"}, 200),
            view_properties={"serializer_class": serializer_class},
        )
        view.method = "get"
        view.format_kwarg = "json"
        view.request = request
        view.action = "retrieve"
        response = view.custom_action(request)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data, {"company_name": "Lenovo"})

    def test_audoma_action_default_collector_fallback(self):
        request = self.factory.post("/custom_action/")
        request.data = self.request_data
        view = create_view_with_custom_audoma_action(
            request=request,
            action_kwargs={
                "results": {201: self.example_result_serializer},
                "methods": ["POST"],
                "detail": False,
            },
            returnables=(self.request_data, 201),
            view_properties={"serializer_class": self.example_collect_serializer},
        )
        view.method = "post"
        view.format_kwarg = "json"
        view.request = request
        view.action = "custom_action"
        response = view.custom_action(request)
        self.assertEqual(response.status_code, 201)
        self.assertEqual(
            response.data, {"firstname": "John", "lastname": "DOG", "age": 45}
        )

    def test_audoma_action_process_defined_exception(self):
        request = self.factory.post("/custom_action/")
        request.data = self.request_data
        view = create_view_with_custom_audoma_action(
            request=request,
            action_kwargs={
                "results": {201: self.example_result_serializer},
                "methods": ["POST"],
                "detail": False,
                "errors": [PermissionDenied("You are not allowed")],
            },
            returnables=(self.request_data, 201),
            view_properties={"serializer_class": self.example_collect_serializer},
            raiseable=PermissionDenied("You are not allowed"),
            raise_exception=True,
        )
        view.method = "post"
        view.request = request
        view.format_kwarg = "json"
        view.action = "custom_action"
        try:
            view.custom_action(request)
        except Exception as e:
            self.assertEqual(type(e), PermissionDenied)
            self.assertEqual(e.detail, "You are not allowed")
            self.assertEqual(e.status_code, 403)

    @override_settings(DEBUG=True)
    def test_audoma_action_process_undefined_exception(self):
        request = self.factory.post("/custom_action/")
        request.data = self.request_data
        view = create_view_with_custom_audoma_action(
            request=request,
            action_kwargs={
                "results": {200: self.example_result_serializer},
                "methods": ["GET"],
                "detail": False,
                "errors": [PermissionDenied("You are not allowed")],
            },
            returnables=(self.request_data, 200),
            view_properties={"serializer_class": self.example_collect_serializer},
            raiseable=MethodNotAllowed("POST"),
            raise_exception=True,
        )
        view.method = "post"
        view.request = request
        view.action = "custom_action"
        view.format_kwarg = "json"
        try:
            view.custom_action(request)
        except Exception as e:
            self.assertEqual(type(e), AudomaActionException)
            self.assertEqual(
                str(e),
                f"There is no class or instance of {MethodNotAllowed} \
                    defined in audoma_action errors.",
            )

    @override_settings(DEBUG=True)
    def test_audoma_action_process_different_content_exception(self):
        request = self.factory.post("/custom_action/")
        request.data = self.request_data

        class CustomException(APIException):
            default_detail = "This does not seem to work"

        raisable = CustomException("Something else happens")
        view = create_view_with_custom_audoma_action(
            request=request,
            action_kwargs={
                "results": {200: self.example_result_serializer},
                "methods": ["GET"],
                "detail": False,
                "errors": [CustomException("Something happens")],
            },
            returnables=(self.request_data, 200),
            view_properties={"serializer_class": self.example_collect_serializer},
            raiseable=raisable,
            raise_exception=True,
        )
        view.method = "post"
        view.request = request
        view.format_kwarg = "json"
        view.action = "custom_action"
        try:
            view.custom_action(request)
        except Exception as e:
            self.assertEqual(type(e), AudomaActionException)
            self.assertEqual(
                str(e),
                f"Raised error: {raisable} has not been \
                        defined in audoma_action errors.",
            )

    def test_audoma_action_many_param_true(self):
        request = self.factory.post("/custom_action/")
        request.data = [self.request_data]

        view = create_view_with_custom_audoma_action(
            request=request,
            action_kwargs={
                "results": {200: self.example_result_serializer},
                "methods": ["GET"],
                "detail": False,
                "many": True,
            },
            returnables=(
                [
                    self.request_data,
                ],
                200,
            ),
            view_properties={"serializer_class": self.example_collect_serializer},
        )
        view.method = "post"
        view.request = request
        view.format_kwarg = "json"
        view.action = "custom_action"
        response = view.custom_action(request)
        self.assertIsInstance(response.data, list)



from django.contrib import admin

from .models import (
    ExampleFileModel,
    ExampleModel,
    ExamplePerson,
)


admin.site.register(ExamplePerson)
admin.site.register(ExampleModel)
admin.site.register(ExampleFileModel)



from rest_framework.exceptions import ValidationError

from django.test import TestCase

from audoma.drf.validators import ExclusiveFieldsValidator


class ExclusiveFieldsValidatorTestCase(TestCase):
    def setUp(self):
        self.validator = ExclusiveFieldsValidator(
            fields=["name", "company_name"],
            message="Fields: {field_names} are mutually exclusive",
            required=True,
            message_required="At least one of the fields: {field_names} is required",
        )

    def test_call_none_in_data(self):
        try:
            self.validator(data={"age": 245})
        except Exception as e:
            self.assertEqual(type(e), ValidationError)
            self.assertEqual(
                str(e.detail[0]),
                "At least one of the fields: name, company_name is required",
            )

    def test_call_all_in_data(self):
        try:
            self.validator(data={"name": "Test", "company_name": "Test"})
        except Exception as e:
            self.assertEqual(type(e), ValidationError)
            self.assertEqual(
                str(e.detail[0]), "Fields: name, company_name are mutually exclusive"
            )

    def test_call_one_in_data(self):
        try:
            val_response = self.validator(data={"name": "Test"})
        except Exception as e:
            raise e
        else:
            self.assertIsNone(val_response)



from functools import wraps
from typing import (
    Any,
    Callable,
)

from drf_spectacular.utils import extend_schema_field
from rest_framework import serializers


def document_and_format(serializer_or_field: Any) -> Callable:
    def decorator(func: Callable) -> Callable:
        @wraps(func)
        @extend_schema_field(serializer_or_field)
        def wrapper(*args, **kwargs) -> Any:
            value = func(*args, **kwargs)
            f = (
                serializer_or_field
                if isinstance(serializer_or_field, serializers.Field)
                else serializer_or_field()
            )
            f.parent = args[0]
            return f.to_representation(value) if value is not None else None

        return wrapper

    return decorator



"""
This module is an override for default drf's field module.
Most of those fields, are providing additional example functionality,
and also has defined schema type.
"""

import sys

import exrex
import phonenumbers
from djmoney.contrib.django_rest_framework import MoneyField
from drf_spectacular.types import OpenApiTypes
from drf_spectacular.utils import extend_schema_field
from phonenumber_field import serializerfields
from phonenumber_field.phonenumber import to_python
from rest_framework import fields
from rest_framework.fields import *  # noqa: F403, F401

from django.core import validators

from audoma.example_generators import generate_lorem_ipsum
from audoma.mixins import (
    ExampleMixin,
    NumericExampleMixin,
    RegexExampleMixin,
)


field_names = [
    "BooleanField",
    "NullBooleanField",
    "EmailField",
    "SlugField",
    "URLField",
    "DateTimeField",
    "DurationField",
    "ChoiceField",
    "MultipleChoiceField",
    "FilePathField",
    "FileField",
    "ImageField",
    "ListField",
    "DictField",
    "HStoreField",
    "JSONField",
    "ReadOnlyField",
    "SerializerMethodField",
]


this = sys.modules[__name__]


for field_name in field_names:

    setattr(
        this,
        field_name,
        type(field_name, (ExampleMixin, getattr(fields, field_name)), {}),
    )


class DecimalField(NumericExampleMixin, fields.DecimalField):
    pass


@extend_schema_field(OpenApiTypes.UUID)
class UUIDField(ExampleMixin, fields.UUIDField):
    pass


class IntegerField(NumericExampleMixin, fields.IntegerField):
    pass


class FloatField(NumericExampleMixin, fields.FloatField):
    pass


class RegexField(RegexExampleMixin, fields.RegexField):
    pass


class MACAddressField(RegexExampleMixin, fields.CharField):
    def __init__(self, **kwargs) -> None:
        self.regex = "^([0-9A-F]{2}:){5}([0-9A-F]{2})|([0-9A-F]{2}-){5}([0-9A-F]{2})$"
        self.validators = [validators.RegexValidator(self.regex)]
        super().__init__(**kwargs)


@extend_schema_field(OpenApiTypes.DATE)
class DateField(ExampleMixin, fields.DateField):
    pass


@extend_schema_field(OpenApiTypes.TIME)
class TimeField(ExampleMixin, fields.TimeField):
    pass


@extend_schema_field(
    field={
        "format": "ip-address",
        "example": str(
            exrex.getone(
                r"^(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$"
            )
        ),
    }
)
class IPAddressField(ExampleMixin, fields.IPAddressField):
    pass


@extend_schema_field(field={"format": "tel"})
class PhoneNumberField(ExampleMixin, serializerfields.PhoneNumberField):
    def __init__(self, *args, **kwargs) -> None:
        example = kwargs.pop("example", None)
        if example is None:
            number = phonenumbers.example_number(None)
            example = str(to_python(number))
        super().__init__(*args, example=example, **kwargs)


class CharField(ExampleMixin, fields.CharField):
    def __init__(self, *args, **kwargs) -> None:
        example = kwargs.pop("example", None)
        min_length = kwargs.get("min_length", 20)
        max_length = kwargs.get("max_length", 80)
        if not example:
            example = generate_lorem_ipsum(min_length=min_length, max_length=max_length)

        super().__init__(*args, example=example, **kwargs)


class MoneyField(NumericExampleMixin, MoneyField):
    pass


class ChoiceField(ExampleMixin, fields.ChoiceField):
    def __init__(self, choices, **kwargs):
        if isinstance(choices, dict):
            self.original_choices = list(choices.items())
        else:
            self.original_choices = choices
        super().__init__(choices, **kwargs)



from inspect import isclass
from typing import (
    Callable,
    List,
    Tuple,
)

from rest_framework.settings import api_settings

from django.conf import settings as project_settings
from django.utils.module_loading import import_string

from audoma import settings as audoma_settings


def preprocess_include_path_format(
    endpoints: Tuple[str, str, str, Callable], **kwargs
) -> List[Tuple[str, str, str, Callable]]:
    """
    Preprocessing hook that filters {format} prefdixed paths, in case
    schema pattern prefix is used and {format} path params are wanted.
    """
    format_path = project_settings.SCHEMA_PATTERN_PREFIX

    return [
        (path, path_regex, method, callback)
        for path, path_regex, method, callback in endpoints
        if (
            path.startswith(format_path)
            or path.startswith(format_path + "/")
            or path.startswith("/" + format_path)
        )
    ]


def postprocess_common_errors_section(result: dict, request, **kwargs) -> dict:
    """
    Postprocessing hook which adds COMMON_API_ERRORS description to the API description.
    """
    common_exceptions = audoma_settings.COMMON_API_ERRORS + getattr(
        project_settings, "COMMON_API_ERRORS", []
    )

    renderer = project_settings.REST_FRAMEWORK.get(
        "DEFAULT_RENDERER_CLASSES", api_settings.DEFAULT_RENDERER_CLASSES
    )[0]

    if not callable(renderer):
        renderer = import_string(renderer)

    renderer = renderer()

    def generate_exception_desc(error):
        exc_desc = ""
        exc_desc = f"Status Code: `{error.status_code}` \n\n"
        rendered_error_data = renderer.render(
            data=vars(error), renderer_context={"indent": 4}
        ).decode("utf-8")

        exc_desc += f"``` \n {rendered_error_data} \n ``` \n\n"
        return exc_desc

    result["info"] = result.get("info", {})

    description = "###  Common API Errors \n"
    for error in common_exceptions:
        if isclass(error):
            error = error()
        description += generate_exception_desc(error)

    result["info"]["description"] = (
        result["info"].get("description", "") + "\n\n" + description
    )

    return result



import random
from itertools import cycle

from django.utils.functional import lazy

from audoma.choices import make_choices
from audoma.django.db import models


# Create your models here.
class ExampleModel(models.Model):

    EXAMPLE_CHOICES = make_choices(
        "CHOICES",
        (
            (1, "EX_1", "example 1"),
            (2, "EX_2", "example 2"),
            (3, "EX_3", "example 3"),
        ),
    )
    char_field = models.CharField(max_length=255)
    phone_number = models.PhoneNumberField()
    phone_number_example = models.PhoneNumberField(example="+123456789")
    email = models.EmailField()
    text_field = models.TextField()
    url = models.URLField()
    boolean = models.BooleanField()
    nullboolean = models.BooleanField(null=True)
    mac_adress = models.MACAddressField()
    slug = models.SlugField()
    uuid = models.UUIDField()
    ip_address = models.GenericIPAddressField()
    integer = models.IntegerField()
    _float = models.FloatField()
    decimal = models.DecimalField(decimal_places=2, max_digits=10)
    datetime = models.DateTimeField()
    date = models.DateField()
    time = models.TimeField()
    duration = models.DurationField()
    choices = models.IntegerField(choices=EXAMPLE_CHOICES.get_choices())
    json = models.JSONField()
    money = models.MoneyField(decimal_places=2, max_digits=10)


example_countries = cycle(["United States", "Canada", "France", "Poland", "Italy"])
example_cities = ["New York", "Athens", "Toronto", "Rome", "Tokyo", "Oslo"]


def get_countries():
    return next(example_countries)


def get_random_age():
    return random.randint(18, 80)


class Country(models.Model):
    name = models.CharField(max_length=225, example=get_countries)


class ExamplePerson(models.Model):
    first_name = models.CharField(max_length=225, example="Adam")
    last_name = models.CharField(max_length=255, example="Smith")
    age = models.IntegerField(example=get_random_age)
    email = models.EmailField(example="example_person@example.com")
    birth_country = models.ForeignKey(Country, on_delete=models.CASCADE)
    residence_city = models.CharField(
        max_length=255, example=lazy(lambda: random.choice(example_cities), str)
    )
    has_valid_account = models.BooleanField()
    ip_address = models.GenericIPAddressField()
    savings = models.MoneyField(max_digits=14, decimal_places=2, default_currency="PLN")
    phone_number = models.PhoneNumberField(region="IT")


class ExampleFileModel(models.Model):
    file_field = models.FileField()
    name = models.CharField(max_length=255)


class Manufacturer(models.Model):
    name = models.CharField(max_length=255)
    slug_name = models.SlugField()


class Car(models.Model):

    CAR_BODY_TYPES = make_choices(
        "BODY_TYPES",
        (
            (1, "SEDAN", "Sedan"),
            (2, "COUPE", "Coupe"),
            (3, "HATCHBACK", "Hatchback"),
            (4, "PICKUP", "Pickup Truck"),
        ),
    )

    ENGINE_TYPES = make_choices(
        "ENGINE_TYPES",
        (
            (1, "PETROL", "Petrol"),
            (2, "DIESEL", "Diesel"),
            (3, "ELECTRIC", "Electric"),
            (4, "HYBRID", "Hybrid"),
        ),
    )

    name = models.CharField(max_length=255)
    body_type = models.IntegerField(choices=CAR_BODY_TYPES.get_choices())

    manufacturer = models.ForeignKey(Manufacturer, on_delete=models.CASCADE)
    engine_size = models.FloatField()
    engine_type = models.IntegerField(choices=ENGINE_TYPES.get_choices())



from dataclasses import dataclass
from inspect import isclass
from typing import (
    Iterable,
    List,
    Type,
    Union,
)

from rest_framework.exceptions import APIException
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.serializers import BaseSerializer

from django.db.models import Model
from django.views import View


@dataclass
class OperationExtractor:

    collectors: Union[dict, Type[BaseSerializer]]
    results: Union[dict, Type[BaseSerializer], str]
    errors: List[Union[APIException, Type[APIException]]]

    def _create_exception(self, options: dict) -> APIException:
        """
        Create proper exception if error status code, found defined in the response.
        Args:
            * options - kwargs for APIException instance.
        Returns: APIException instance.
        """
        status_code = options.pop("status_code")
        exception = APIException(**options)
        exception.status_code = status_code if status_code else exception.status_code
        return exception

    def _extract_response_operation(
        self, request: Request, code: int
    ) -> Union[dict, Type[BaseSerializer], str, APIException]:
        """
        Extracts response operation from the defined in audoma_action dictionary.

        Args:
            * request - request object
            * code - response status code, this should be passed as an integer.
        Returns: Serializer instance, dictionary, string or APIException.
        """
        if code >= 400:
            error_data = self.errors.get(code, {})
            error_kwargs = {
                "status_code": code,
                "detail": error_data.get("detail"),
                "code": error_data.get("error_code"),
            }
            return self._create_exception(error_kwargs)

        if not self.results or (
            isclass(self.results) and issubclass(self.results, BaseSerializer)
        ):
            return self.results

        if isinstance(list(self.results.keys())[0], str) and request:
            method = request.method.lower()
            response = self.results.get(method)
            if code and isinstance(response, dict):
                response = response.get(code)
            return response

        elif isinstance(list(self.results.keys())[0], int) and code:
            response = self.results.get(code)
            return response

        return None

    def _extract_collect_operation(self, request: Request) -> Type[BaseSerializer]:
        """
        Extracts collect operation from the defined in audoma_action dictionary.

        Args:
            * request - request object
        Returns: Serializer class.
        """
        if not self.collectors or (
            isclass(self.collectors) and issubclass(self.collectors, BaseSerializer)
        ):
            return self.collectors

        method = request.method.lower()
        return self.collectors.get(method)

    def extract_operation(
        self, request: Request, code: int = None, operation_category="response"
    ) -> Union[dict, Type[BaseSerializer], str, APIException]:
        """
        Extracts proper operation from the defined in audoma_action dictionary.
        Args:
            * request - request object
            * code - response status code, this should be given only for responses.
                It should be given as an integer
            * operation_category - operation category, it can be either "response" or "collect".
                This determines if the extracted operation should be either response or request operation.
        Returns: Serializer instance, dictionary, string or APIException.
        """
        if operation_category == "response":
            return self._extract_response_operation(request, code)
        elif operation_category == "collect":
            return self._extract_collect_operation(request)
        else:
            raise ValueError("Unknown operation_category")


def apply_response_operation(
    operation: Union[str, APIException, Type[BaseSerializer]],
    instance: Union[Iterable, str, APIException, Model],
    code: int,
    view: View,
    many: bool,
) -> Response:
    """
    Applies response operation for automa_action decorator.
    Args:
        * operation: response operation wich will be applied, it may be a string message, APIException object, or
            resoponse serializer_class instance
        * instance: instance which will be returned in a response. It may be iterable, string, ApiException object,
            or simply model instance. If it's dict/model instance it'll be serialized.
        * code - response status code, this should be given as an integer
        * view - view instance for which the response will be returned.
        * many - boolean value which determines whether the serializer
            will handle multiple instances, or just singular instnace.

    Returns: Response object.
    """
    if isinstance(operation, APIException):
        raise operation

    if isinstance(operation, str):
        instance = instance or operation
        instance = {"message": instance}
        return Response(instance, status=code)

    serializer_class = operation
    serializer_kwargs = {
        "instance": instance,
        "many": many,
        "context": {"request": view.request, "format": view.format_kwarg, "view": view},
    }
    return_serializer = (
        serializer_class(**serializer_kwargs)
        if serializer_class
        else view.get_result_serializer(**serializer_kwargs)
    )

    headers = view.get_success_headers(return_serializer.data)
    return Response(return_serializer.data, status=code, headers=headers)



from typing import (
    Any,
    List,
    Tuple,
    TypeVar,
)


_T = TypeVar("_T")


def make_choices(name: str, choices_tuple: Tuple[Any, str, str]) -> _T:
    from collections import namedtuple

    """Factory function for quickly making a namedtuple suitable for use in a
    Django model as a choices attribute on a field. It will preserve order.

    Usage::

        class MyModel(models.Model):
            COLORS = make_choices('COLORS', (
                (0, 'BLACK', 'Black'),
                (1, 'WHITE', 'White'),
            ))
            colors = models.PositiveIntegerField(choices=COLORS)

        >>> MyModel.COLORS.BLACK
        0
        >>> MyModel.COLORS.get_choices()
        [(0, 'Black'), (1, 'White')]

        class OtherModel(models.Model):
            GRADES = make_choices('GRADES', (
                ('FR', 'FR', 'Freshman'),
                ('SR', 'SR', 'Senior'),
            ))
            grade = models.CharField(max_length=2, choices=GRADES)

        >>> OtherModel.GRADES.FR
        'FR'
        >>> OtherModel.GRADES.get_choices()
        [('FR', 'Freshman'), ('SR', 'Senior')]

    """

    class Choices(namedtuple(name, [name_ for val, name_, desc in choices_tuple])):
        __slots__ = ()
        _choices = tuple([desc for val, name_, desc in choices_tuple])

        def get_display(self, val: Any) -> str:
            return self._choices[self.index(val)]

        def get_choices(self) -> List[Tuple[Any, str]]:
            return list(zip(tuple(self), self._choices))

        def get_api_choices(self) -> List[Tuple[str, str]]:
            return list(zip(tuple(self._asdict()), tuple(self._choices)))

        def get_value_by_name(self, name: str) -> Any:
            return getattr(self, name)

    return Choices._make([val for val, name_, desc in choices_tuple])



from typing import (
    List,
    Tuple,
    Type,
    Union,
)

from drf_spectacular.drainage import cache
from rest_framework import (
    generics,
    mixins,
    serializers,
    views,
    viewsets,
)

from audoma.drf import (
    generics as audoma_generics,
    mixins as audoma_mixins,
    serializers as audoma_serializers,
    viewsets as audoma_viewsets,
)


@cache
def get_lib_doc_excludes_audoma() -> List[Type]:
    return [
        views.APIView,
        *[
            getattr(serializers, c)
            for c in dir(serializers)
            if c.endswith("Serializer")
        ],
        *[getattr(viewsets, c) for c in dir(viewsets) if c.endswith("ViewSet")],
        *[getattr(generics, c) for c in dir(generics) if c.endswith("APIView")],
        *[getattr(mixins, c) for c in dir(mixins) if c.endswith("Mixin")],
        *[
            getattr(audoma_viewsets, c)
            for c in dir(audoma_viewsets)
            if c.endswith("ViewSet")
        ],
        *[getattr(audoma_mixins, c) for c in dir(audoma_mixins) if c.endswith("Mixin")],
        *[
            getattr(audoma_serializers, c)
            for c in dir(audoma_serializers)
            if c.endswith("Serializer")
        ],
        *[
            getattr(audoma_generics, c)
            for c in dir(audoma_generics)
            if c.endswith("APIView")
        ],
    ]


def create_choices_enum_description(choices: Union[dict, List[Tuple]], field_name):

    if not isinstance(choices, dict):
        choices = dict(choices)

    description = f"Filter by {field_name} \n"
    for key, val in choices.items():
        description += f" * `{key}` - {val}\n"
    return description



# Generated by Django 4.0.4 on 2022-05-19 06:10

import django.db.models.deletion
from django.db import (
    migrations,
    models,
)

import audoma.django.db.fields


class Migration(migrations.Migration):

    dependencies = [
        ("audoma_api", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="ExampleSimpleModel",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", audoma.django.db.fields.CharField(max_length=64)),
                ("value", audoma.django.db.fields.IntegerField(default=0, null=True)),
            ],
        ),
        migrations.AlterField(
            model_name="examplemodel",
            name="json",
            field=audoma.django.db.fields.JSONField(),
        ),
        migrations.AlterField(
            model_name="examplemodel",
            name="money_currency",
            field=audoma.django.db.fields.CurrencyField(
                choices=[
                    ("AUD", "Australian Dollar"),
                    ("GBP", "British Pound"),
                    ("CNY", "Chinese Yuan"),
                    ("EUR", "Euro"),
                    ("INR", "Indian Rupee"),
                    ("JPY", "Japanese Yen"),
                    ("NZD", "New Zealand Dollar"),
                    ("CHF", "Swiss Franc"),
                    ("USD", "US Dollar"),
                ],
                default="XYZ",
                max_length=3,
            ),
        ),
        migrations.AlterField(
            model_name="exampleperson",
            name="savings_currency",
            field=audoma.django.db.fields.CurrencyField(
                choices=[
                    ("AUD", "Australian Dollar"),
                    ("GBP", "British Pound"),
                    ("CNY", "Chinese Yuan"),
                    ("EUR", "Euro"),
                    ("INR", "Indian Rupee"),
                    ("JPY", "Japanese Yen"),
                    ("NZD", "New Zealand Dollar"),
                    ("CHF", "Swiss Franc"),
                    ("USD", "US Dollar"),
                ],
                default="PLN",
                max_length=3,
            ),
        ),
        migrations.CreateModel(
            name="ExampleTagModel",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", audoma.django.db.fields.CharField(max_length=32)),
                (
                    "item",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="tags",
                        to="audoma_api.examplesimplemodel",
                    ),
                ),
            ],
        ),
    ]
